import 'dart:convert';
import 'package:http/http.dart' as http;
import 'package:modular_system_core/mscore.dart';
import 'package:modular_system_core/src/model/identity_credential.dart';

class WebController {
  WebController({
    required this.baseUri,
    required this.controllerName,
  });

  final String baseUri;
  final String controllerName;
  late final String fullUri = "$baseUri/$controllerName";

  Map<String, dynamic> getJsonFrom(Response resp) {
    if (resp.statusCode == 200) {
      return jsonDecode(resp.body);
    } else if (resp.body.isNotEmpty) {
      throw AppException.fromJson(resp.body);
    } else {
      throw AppException("${resp.statusCode} ${resp.reasonPhrase}");
    }
  }

  void handleResponse(Response resp) {
    if (resp.statusCode == 200) {
      return;
    } else if (resp.body.isNotEmpty) {
      throw AppException.fromJson(resp.body);
    } else {
      throw AppException("${resp.statusCode} ${resp.reasonPhrase}");
    }
  }
}

class ApiController<T extends JsonSerializer<T>> extends WebController {
  ApiController({
    required String baseUri,
    required String controllerName,
    required this.objectCreator,
  }) : super(baseUri: baseUri, controllerName: controllerName);

  final JsonObjectCreator<T> objectCreator;

  Future<StringDTO> createAsync(T object) async {
    var resp = await http.post(
      Uri.parse(fullUri),
      headers: <String, String>{
        'Content-Type': 'application/json',
      },
      body: object.toJson(),
    );
    var jsonMap = getJsonFrom(resp);
    return StringDTO.fromJson(jsonMap);
  }

  // READ
  Future getByIdAsync(String id) async {
    var resp = await http.get(
      Uri.parse("$fullUri?id=$id"),
      headers: <String, String>{
        'Content-Type': 'application/json; charset=UTF-8',
      },
    );
    var jsonMap = getJsonFrom(resp);
    return objectCreator(jsonMap);
  }

  Future<QueryList<T>> searchAsync(WebSearchRequest request) async {
    var resp = await http.post(
      Uri.parse("$fullUri/query"),
      headers: <String, String>{
        'Content-Type': 'application/json;',
        'Access-Control-Allow-Origin': '*'
      },
      body: utf8.encode(request.toJson()),
    );
    var jsonMap = getJsonFrom(resp);
    return QueryList.fromJsonMap(jsonMap, objectCreator);
  }

  // UPDATE
  Future updateAsync(WebUpdateRequest request) async {
    var resp = await http.put(
      Uri.parse(fullUri),
      headers: <String, String>{
        'Content-Type': 'application/json; charset=UTF-8',
        'Access-Control-Allow-Origin': '*'
      },
      body: request.toJson(),
    );
    handleResponse(resp);
  }

  Future overrideAsync(T request) async {
    var resp = await http.put(
      Uri.parse("$fullUri/override"),
      headers: <String, String>{
        'Content-Type': 'application/json; charset=UTF-8',
        'Access-Control-Allow-Origin': '*'
      },
      body: request.toJson(),
    );
    handleResponse(resp);
  }

  // DELETE
  Future deleteAsync(String id) async {
    var resp = await http.delete(
      Uri.parse("$fullUri?id=$id"),
      headers: <String, String>{
        'Content-Type': 'application/json',
      },
    );
    handleResponse(resp);
  }
}

class AuthenticationController extends WebController {
  AuthenticationController({
    required String baseUri,
    required String controllerName,
  }) : super(baseUri: baseUri, controllerName: controllerName);

  Future<String> authenticateAsync(CredentialHandshake handshake) async {
    var resp = await http.post(
      Uri.parse("$fullUri/authenticate"),
      headers: <String, String>{
        'Content-Type': 'application/json',
      },
      body: handshake.toJson(),
    );
    var jsonMap = getJsonFrom(resp);
    return StringDTO.fromJson(jsonMap).value;
  }
}
